#############
# $Id: Checker.pm 13239 2011-05-23 08:39:19Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2011-02-16
# Last Changed On: $Date: 2011-05-23 09:39:19 +0100 (Mon, 23 May 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-29.0/lib/npg_pipeline/Checker.pm $

package npg_pipeline::Checker;
use Moose;
use Carp;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13239 $ =~ /(\d+)/mxs; $r; };

extends qw{npg_pipeline::base};

=head1 NAME

npg_pipeline::Checker

=head1 VERSION

$LastChangedRevision: 13239 $

=head1 SYNOPSIS

  use npg_pipeline::Checker;

  npg_pipeline::Checker->new()->check();

=head1 DESCRIPTION

This object module looks at the LSF statuses of jobs, and tries to work out if they are stuck, and what it causing them to be so.

=head1 SUBROUTINES/METHODS

=head2 check

Method to run, it does everything. The output will we produced to STDOUT. Optionally, you may add a single id_run, and it will
only look at that one, outputting nothing if there are no problems.

  npg_pipeline::Checker->new()->check();
  npg_pipeline::Checker->new()->check( $id_run );

=cut


sub check {
  my ( $self, $id_run ) = @_;

  my $stuck_runs = $self->stuck_runs();

  if ( $id_run ) {
    my $wanted_run;
    foreach my $stuck_run ( @{ $stuck_runs } ) {
      if ( $stuck_run->{id_run} == $id_run ) {
        $wanted_run = $stuck_run;
        last;
      }
    }
    if ( ! $wanted_run ) {
      return 1;
    }
    $self->_overwrite_stuck_runs( [ $wanted_run ] );
    $stuck_runs = $self->stuck_runs();
  }

  foreach my $run ( @{ $stuck_runs } ) {
    $self->_stuck_run_information( $run );
  }

  return 1;
}


has q{stuck_runs} => (
  isa => q{ArrayRef[HashRef]},
  is  => q{ro},
  init_arg => undef,
  lazy_build => 1,
  writer => q{_overwrite_stuck_runs},
);


sub _build_stuck_runs {
  my ( $self ) = @_;

  my $command = q{bjobs -u prod_users -wp | /software/perl-5.10.1/bin/perl -E 'my$l; sub p {if (defined $l){$l=~s/\n//g; say $l} undef $l} while(<>){p if (/^\S/); $l.=$_} p' | grep invalid | grep -v PB_ | perl -nle 'print $1 if /_0?(\d{4,5})\D/' | sort | uniq | perl -nle 'use npg::api::run; my $r = npg::api::run->new({id_run=>$_}); print "$_ ".$r->current_run_status->description;'}; ## no critic (ValuesAndExpressions::RequireInterpolationOfMetachars)

  my $output = $self->run_command()->( $command, q{'get stuck_runs'} );

  my @runs = split /\n/xms, $output;
  my $return_runs = [];
  foreach my $run ( @runs ) {
    my ( $id_run, $status ) = $run =~ /\A(\d+)[ ](.*)/xms;
    push @{ $return_runs }, { id_run => $id_run, status => $status };
  }

  return $return_runs;
}

sub _stuck_run_information {
  my ( $self, $run ) = @_;
  my $id_run = $run->{id_run};
  print qq{$id_run $run->{status}\n} or $self->log( qq{$id_run $run->{status}} );

  my @stuck_jobs = $self->_stuck_jobs_per_run( $id_run );

  # we'll assume that, regardless of the number of different jobs there are, the first one will show all the dependencies
  # that are relevant.

  my $stuck_job = shift @stuck_jobs;
  @stuck_jobs = (); # probably unnecessary, but there could be a lot of jobs pending if the runs are heavily multiplexed

  my ( $job_id ) = $stuck_job =~ /\A(\d+)/xms;
  my @job_dependencies = $self->_get_job_dependencies( $job_id );

  foreach my $job_dependency_id ( @job_dependencies ) {
    $self->_assess_job_dependency( $job_dependency_id );
  }

  return 1;
}

sub _stuck_jobs_per_run {
  my ( $self, $id_run ) = @_;

  my $command = q{bjobs -u prod_users -wp | /software/perl-5.10.1/bin/perl -E 'my$l; sub p {if (defined $l){$l=~s/\n//g; say $l} undef $l} while(<>){p if (/^\S/); $l.=$_} p' | grep invalid | grep -E '_0{0,1}} . $id_run . q{_'}; ## no critic (ValuesAndExpressions::RequireInterpolationOfMetachars)

  my $output = $self->run_command()->( $command, qq{'get run_information $id_run'} );

  my @stuck_jobs = split /\n/xms, $output;

  return @stuck_jobs;
}

sub _assess_job_dependency {
  my ( $self, $job_id ) = @_;

  my $command = qq{bhist -l $job_id};
  my $output = $self->run_command()->( $command, qq{bhist $job_id} );

  if ( $output !~ /Exited|Termination|Killed/xms ) {
    return 1;
  }

  my @sections = split /------------------------------------------------------------------------------/xms, $output;

  foreach my $section ( @sections ) {
    next if $section !~ /Exited|Termination|Killed/xms;

    my ( $job_name_and_id ) = $section =~ /(Job[ ]<.*?>,[ ]Job[ ]Name[ ]<.*?>),/xms;
    $job_name_and_id =~ s/\s+/ /xms;
    print qq{\t$job_name_and_id\n} or $self->log( qq{\t$job_name_and_id} );
  }
  return 1;
}

sub _get_job_dependencies {
  my ( $self, $job_id ) = @_;
  my $command = qq{bjobs -l $job_id};
  my $output = $self->run_command()->( $command, qq{bjobs $job_id} );

  ( $output ) = $output =~ /\A(.*?)PENDING[ ]REASONS/xms;
  $output =~ s/\s//gxms;
  ( $output ) = $output =~ /DependencyCondition<(.*?)>/ixms;

  my @deps = split /&&/xms, $output;
  my @dep_jids;
  foreach my $dep ( @deps ) {
    my ( $jid ) = $dep =~ /(\d+)/xms;
    push @dep_jids, $jid;
  }
  return @dep_jids;
}

# run_command has been set as a code_ref, so it can be overwritten in tests
has q{run_command} => (
  isa => q{CodeRef},
  is  => q{ro},
  lazy_build => 1,
);

sub _build_run_command {
  my ( $self ) = @_;
  my $return = sub {
    my ( $command, $command_name ) = @_;

    my $output = q{};
    eval {
      $output = qx{$command};
      if ( $CHILD_ERROR ) {
        $output = q{};
        croak qq{Error running command $command_name: $output};
      }
      1;
    } or do {
      carp $EVAL_ERROR;
    };

    return $output;
  };
  return $return;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2011 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
