#############
# $Id: file_constructs.pm 13564 2011-07-06 13:42:20Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2009-11-04
# Last Changed On: $Date: 2011-07-06 14:42:20 +0100 (Wed, 06 Jul 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-29.0/lib/npg_pipeline/roles/business/file_constructs.pm $

package npg_pipeline::roles::business::file_constructs;
use Moose::Role;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13564 $ =~ /(\d+)/mxs; $r; };

requires qw{id_run use_bases is_indexed};

# put the constructs as global vars here, as they are done so the whole pipeline will always use and expect them
Readonly::Scalar our $FASTQ_FILE_CONSTRUCTS       => q{substitute_this.fastq};
Readonly::Scalar our $FASTQCHECK_FILE_CONSTRUCTS  => q{substitute_this.fastqcheck};
Readonly::Scalar our $MD5_FILE_CONSTRUCTS         => q{substitute_this.md5};
Readonly::Scalar our $QSEQ_FILE_CONSTRUCTS        => q{s_substitute_this_*_qseq_custom.txt};

=head1 NAME

npg_pipeline::roles::business::file_constructs

=head1 VERSION

$LastChangedRevision: 13564 $

=head1 SYNOPSIS

  package MyPackage;
  use Moose;
  ...
  with qw{npg_pipeline::roles::business::file_constructs};

=head1 DESCRIPTION

This role provides the internal workings to generate file_names with a 'substitute_this' part for replacement.
An id_run method is required to provide the id_run part to the name

=head1 SUBROUTINES/METHODS

=head2 fastq_file_constructs
=head2 fastqcheck_file_constructs
=head2 md5_file_constructs

Each of these returns a string of the filename with a 'substitute_this' part to switch out to make filenames unique.
Currently they will all return as follows

  <id_run>_substitute_this.<extension>

e.g.

  1234_substitute_this.fastq

=cut

sub fastq_file_constructs {
  my ($self) = @_;
  return $self->id_run() . q{_} . $FASTQ_FILE_CONSTRUCTS;
}

sub fastqcheck_file_constructs {
  my ($self) = @_;
  return $self->id_run() . q{_} . $FASTQCHECK_FILE_CONSTRUCTS;
}

sub md5_file_constructs {
  my ($self) = @_;
  return $self->id_run() . q{_} . $MD5_FILE_CONSTRUCTS;
}

=head2 qseq_file_constructs - This returns an arrayref of qseq_files with the path they ought to be found at dependent on if
they are indexed reads or not. For this, you need to provide the lane position number, or the string 'all_lanes'. These files
are found from the recalibrated_path and you must have a method exposing recalibrated_path and tag_qseqs_path

  my $aQseqFileConstructs = $class->qseq_file_constructs();

=cut

sub qseq_file_constructs {
  my ($self, $file_name) = @_;

  my $qseq = $QSEQ_FILE_CONSTRUCTS;
  my (@file_names, @files);

  if ($file_name eq q{all_lanes}) {

    my $run = $self->run();
    if ($run) {
      foreach my $lane (@{$run->run_lanes()}) {
        push @file_names, $lane->position();
      }
    } else {
      @file_names = (qw{1 2 3 4 5 6 7 8});
    }

  } else {
    push @file_names, $file_name;
  }

  my @dir_files;

  eval {
    my $dh;
    opendir $dh, $self->recalibrated_path();
    @dir_files = readdir $dh;
    closedir $dh;
  } or do {
    croak $EVAL_ERROR
  };

  foreach my $fn (@file_names) {

    my $qseq_file_name = $QSEQ_FILE_CONSTRUCTS;

    my $head_replace_if_array;
    my $lsb_jobindex = q{LSB_JOBINDEX};

    if ( $file_name =~ /${lsb_jobindex}/xms ) {

      my $temp = $QSEQ_FILE_CONSTRUCTS;
      $temp =~ s/substitute_this/$file_name/xms;

      ($head_replace_if_array, $temp) = split /\*/xms, $temp; ## no critic (RegularExpressions::ProhibitEscapedMetacharacters)

      # just grab for first lane. We'll assume for now that all lanes are equal in number of tiles
      if ( $file_name =~ /_(\d+)/xms ) {
        my @positions = $self->positions();
        $fn = $positions[0] . q{_} . $1;
      } else {
        $fn = 1;
      }
    }

    $qseq_file_name =~ s/substitute_this/$fn/xms;

    my ($head,$tail) = split /\*/xms, $qseq_file_name; ## no critic (RegularExpressions::ProhibitEscapedMetacharacters)


    foreach my $f (@dir_files) {

      if (
          $f =~ /$head/xms
            &&
          $f =~ /\d+$tail\z/xms
            &&
          $f !~ /_t_/xms
         ) {
        if ( $head_replace_if_array ) {
          $f =~ s/$head/$head_replace_if_array/xms;
        }
        push @files, $self->recalibrated_path() . q{/} . $f;
      }

    }

  }

  my @sorted_files = sort @files;
  my $return_string = join q{ }, @sorted_files;
  $return_string ||= q{};

  return $return_string;
}

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose::Role

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
