# $Id: util.pm 13883 2011-08-10 09:55:39Z dj3 $
package t::util;

use Moose;
use Carp;
use English qw{-no_match_vars};
use YAML qw(LoadFile);
use MIME::Parser;
use MIME::Lite;
use Test::More;
use t::useragent;
use npg::api::util;
use File::Temp qw{ tempdir };
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13883 $ =~ /(\d+)/mx; $r; };
use Cwd;
use npg::api::request;

Readonly::Scalar our $TEMP_DIR => q{/tmp};
Readonly::Scalar our $NFS_STAGING_DISK => q{/nfs/sf45};

has q{emails} => (isa => q{ArrayRef}, is => q{ro}, lazy_build => 1, reader => q{get_emails});
sub _build_emails {
  return [];
}

has q{parsed_email} => (isa => q{HashRef}, is => q{ro}, reader => q{get_parsed_email}, writer => q{set_parsed_email});

has q{cwd} => (
  isa => q{Str},
  is => q{ro},
  lazy_build => 1,
);

sub _build_cwd {
  my ( $self ) = @_;
  return getcwd();
}

###########
# for catching emails, so firstly they don't get sent from within a test
# and secondly you could then parse the caught email
#

sub catch_email {
  my ($self) = @_;

  my $sub = sub {
              my $msg = shift;
      	      push @{$self->get_emails()}, $msg->as_string;
              return;
        };
  MIME::Lite->send('sub', $sub);
  return;
}

##########
# for parsing emails to get information from them, probably caught emails
#

sub parse_email {
  my ($self, $email) = @_;
  if (!$email) {
    $email = shift @{$self->get_emails()};
  }

  my $parser = MIME::Parser->new();
  $parser->output_to_core(1);
  my $entity = $parser->parse_data($email);
  my $ref    = {
	    body          => $entity->bodyhandle->as_string()     || undef,
	    subject       => $entity->head->get('Subject', 0)     || undef,
	    to            => $entity->head->get('To',0)           || undef,
	    cc            => $entity->head->get('Cc',0)           || undef,
	    bcc           => $entity->head->get('Bcc',0)          || undef,
	    from          => $entity->head->get('From',0)         || undef,
	    precendence   => $entity->head->get('Precedence',0)   || undef,
	    content_type  => $entity->head->get('Content-Type',0) || undef,
    };

  return $ref;
}

has q{apiutil} => (isa => q{Any}, is => q{rw}, lazy_build => 1);

sub _build_apiutil {
  my ($self) = @_;
  return npg::api::util->new({
        useragent => $self->useragent(),
      });
}

has q{useragent} => (isa => q{Any}, is => q{rw}, lazy_build => 1);

sub _build_useragent {
  my ($self) = @_;
  return t::useragent->new({
        is_success => 1,
        root_of_t_directory => $self->cwd(),
      });
}

sub serialized_apiutil {
  my ( $self ) = @_;
  use Data::Dumper;
  return Dumper ( $self->apiutil() );
}

sub serialized_useragent {
  my ( $self ) = @_;
  return Dumper ( $self->useragent() );
}

# for getting a temporary directory which will clean up itself, and should not clash with other people attempting to run the tests

has q{temp_directory} => (
  isa => q{Str},
  is => q{ro},
  lazy_build => 1,
);

sub _build_temp_directory {
  my ( $self ) = @_;
  $ENV{TEST_FS_RESOURCE} = q{nfs_12};

  my $tempdir = tempdir(
    DIR => $TEMP_DIR,
    CLEANUP => 1,
  );

  note $tempdir;
  return $tempdir;
}


###############
# path setups

Readonly::Scalar our $DEFAULT_RUNFOLDER => q{123456_IL2_1234};

Readonly::Scalar our $ANALYSIS_RUNFOLDER_PATH => $NFS_STAGING_DISK . q{/IL2/analysis/} . $DEFAULT_RUNFOLDER;
Readonly::Scalar our $OUTGOING_RUNFOLDER_PATH => $NFS_STAGING_DISK . q{/IL2/outgoing/} . $DEFAULT_RUNFOLDER;
Readonly::Scalar our $BUSTARD_PATH            => qq{$ANALYSIS_RUNFOLDER_PATH/Data/C1-18-Firecrest1.3.4_09-07-2009_auto/Bustard1.3.4_09-07-2009_auto};
Readonly::Scalar our $RECALIBRATED_PATH       => qq{$BUSTARD_PATH/GERALD_09-07-2009_auto};
Readonly::Scalar our $REFERENCES_PATH         => q{/nfs/repository/d0031/references};

sub default_runfolder {
  my ( $self ) = @_;
  return $DEFAULT_RUNFOLDER;
}

sub create_references_directory {
  my ( $self ) = @_;
  my $ref_dir = $self->temp_directory() . $REFERENCES_PATH;
  qx{mkdir -p $ref_dir};
  return $ref_dir;
}

sub test_id_run {
  my ($self) = @_;
  my ($id_run) = $self->temp_directory() . $ANALYSIS_RUNFOLDER_PATH =~ /IL\d+_(\d+)/xms;
  return $id_run;
}

sub test_run_folder {
  my ($self) = @_;
  my $test_run_folder_path = $self->temp_directory() . $ANALYSIS_RUNFOLDER_PATH;
  my ($run_folder) = $test_run_folder_path =~ /(\d+_IL\d+_\d+)/xms;
  return $run_folder;
}

sub analysis_runfolder_path {
  my ( $self ) = @_;
  return $self->temp_directory() . $ANALYSIS_RUNFOLDER_PATH;
}

sub standard_analysis_bustard_path {
  my ( $self ) = @_;
  return $self->temp_directory() . $BUSTARD_PATH;
}

sub standard_analysis_recalibrated_path {
  my ( $self ) = @_;
  return $self->temp_directory() . $RECALIBRATED_PATH;
}

sub outgoing_runfolder_path {
  my ( $self ) = @_;
  return $self->temp_directory() . $OUTGOING_RUNFOLDER_PATH;
}

sub set_single_analysis_area {
  my ($self, $args) = @_;
  my $analysis_runfolder_path = $self->temp_directory() . $ANALYSIS_RUNFOLDER_PATH;
  $args ||= {};
  $self->remove_staging();
  `mkdir -p $analysis_runfolder_path/Config`;
  `mkdir -p $analysis_runfolder_path/Data/Intensities/Bustard1.3.4_09-07-2009_single/GERALD_09-07-2009_single`;
  `touch $analysis_runfolder_path/Data/Intensities/Bustard1.3.4_09-07-2009_single/GERALD_09-07-2009_single/touch_file`;
  return 1;
}

sub create_analysis {
  my ($self, $args) = @_;
  $args ||= {};
  $self->remove_staging();
  my $analysis_runfolder_path = $self->temp_directory() . $ANALYSIS_RUNFOLDER_PATH;
  my $recalibrated_path = $self->temp_directory() . $RECALIBRATED_PATH;
  `mkdir -p $recalibrated_path`;
  if (!$args->{skip_archive_dir}) {
    `mkdir -p $recalibrated_path/archive/log`;
  }
  if ($args->{qc_dir}) {
    `mkdir -p $recalibrated_path/archive/log`;
    `mkdir -p $recalibrated_path/archive/qc/log`;
  }
  `mkdir $analysis_runfolder_path/Config`;
  `cp t/data/Recipes/Recipe_GA2_37Cycle_PE_v6.1.xml $analysis_runfolder_path/`;
  `cp t/data/Recipes/TileLayout.xml $analysis_runfolder_path/Config/`;
  `mkdir $analysis_runfolder_path/Data/C1-18-Firecrest1.3.4_09-07-2009_auto/archive`;
  `ln -s Data/C1-18-Firecrest1.3.4_09-07-2009_auto/Bustard1.3.4_09-07-2009_auto/GERALD_09-07-2009_auto $analysis_runfolder_path/Latest_Summary`;
  if (!$args->{no_touch_files}) {
    `touch $recalibrated_path/s_1_1_qseq_custom.txt`;
    `touch $recalibrated_path/s_2_1_qseq_custom.txt`;
    `touch $recalibrated_path/s_1_2_qseq_custom.txt`;
    `touch $recalibrated_path/s_3_1_qseq_custom.txt`;
    `touch $recalibrated_path/s_4_1_qseq_custom.txt`;
    `touch $recalibrated_path/s_5_1_qseq_custom.txt`;
    `touch $recalibrated_path/s_6_1_qseq_custom.txt`;
    `touch $recalibrated_path/s_7_1_qseq_custom.txt`;
    `touch $recalibrated_path/s_8_1_qseq_custom.txt`;
  }
  if ( $args->{bam_generation} ) {
    `touch $recalibrated_path/s_1_1_0001_qseq_custom.txt`;
    `touch $recalibrated_path/../s_1_1_0001_qseq.txt`;

    `touch $recalibrated_path/../s_2_1_0001_qseq.txt`;
    `ln -s  $recalibrated_path/../s_2_1_0001_qseq.txt $recalibrated_path/s_2_1_0001_qseq_custom.txt`;
  }
  return 1;
}

sub create_multiplex_analysis {
  my ($self, $args) = @_;
  $args ||= {};

  $self->create_analysis($args);
  my $analysis_runfolder_path = $self->temp_directory() . $ANALYSIS_RUNFOLDER_PATH;
  my $recalibrated_path = $self->temp_directory() . $RECALIBRATED_PATH;
  `rm $analysis_runfolder_path/Recipe_GA2_37Cycle_PE_v6.1.xml`;
  `cp t/data/Recipes/Recipe_GA2-PEM_MP_2x76Cycle+8_v7.7.xml $analysis_runfolder_path/`;

  if ($args->{qc_dir}) {
    foreach my $lane (@{$args->{qc_dir}}) {
      `mkdir -p $recalibrated_path/archive/lane$lane/qc`;
    }
  }
  if (!$args->{no_touch_files}) {
    `touch $recalibrated_path/s_1_3_qseq_custom.txt`;
  }

  return 1;
}

sub copy_tag_file{
   my ($self, $lane) = @_;
   my $analysis_runfolder_path = $self->temp_directory() . $ANALYSIS_RUNFOLDER_PATH;
   `cp t/data/tag_files/lane_$lane.tag $analysis_runfolder_path`;
}

sub create_tag_qseq {
  my ($self) = @_;
  $self->create_analysis({no_touch_files => 1});
  my $recalibrated_path = $self->temp_directory() . $RECALIBRATED_PATH;
  `mkdir $recalibrated_path/tag_qseqs`;
  foreach my $lane (1..8) {
    foreach my $read (1..3) {
      my $qseq_custom_file = qq{s_${lane}_${read}_1_qseq_custom.txt};
      `touch $recalibrated_path/$qseq_custom_file`;
      `touch $recalibrated_path/tag_qseqs/$qseq_custom_file`;
    }
  }
  return 1;
}


sub set_staging_analysis_area {
  my ($self, $args) = @_;
  $args ||= {};
  $self->remove_staging();
  my $analysis_runfolder_path = $self->temp_directory() . $ANALYSIS_RUNFOLDER_PATH;
  my $bustard_path = $self->temp_directory() . $BUSTARD_PATH;
  my $recalibrated_path = $self->temp_directory() . $RECALIBRATED_PATH;
  `mkdir -p $recalibrated_path`;
  `mkdir $analysis_runfolder_path/Config`;
  `mkdir $analysis_runfolder_path/t`;
  `touch $bustard_path/s_1_1_001_qseq.txt`;
  `touch $bustard_path/s_1_2_001_qseq.txt`;
  `touch $bustard_path/s_2_1_001_qseq.txt`;
  `touch $bustard_path/s_2_2_001_qseq.txt`;
  foreach my $lane (1..8) {
    `touch $recalibrated_path/s_${lane}_1_001_qseq_custom.txt`;
    `touch $recalibrated_path/s_${lane}_2_001_qseq_custom.txt`;
  }
  `cp t/data/Recipes/Recipe_GA2_37Cycle_PE_v6.1.xml $analysis_runfolder_path/`;
  `cp t/data/Recipes/TileLayout.xml $analysis_runfolder_path/Config/`;
  `touch $recalibrated_path/touch_file`;
  if ($args->{with_latest_summary}) {
    `ln -s Data/C1-18-Firecrest1.3.4_09-07-2009_auto/Bustard1.3.4_09-07-2009_auto/GERALD_09-07-2009_auto $analysis_runfolder_path/Latest_Summary`;
  }
  return 1;
}

sub set_staging_outgoing_area {
  my ($self) = @_;
  my $outgoing_runfolder_path = $self->temp_directory() . $OUTGOING_RUNFOLDER_PATH;
  $self->remove_staging();
  `mkdir -p $outgoing_runfolder_path/Data/C1-18-Firecrest1.3.4_09-07-2009_auto/Bustard1.3.4_09-07-2009_auto/GERALD_09-07-2009_auto`;
  `mkdir $outgoing_runfolder_path/Config`;
  `cp t/data/Recipes/Recipe_GA2_37Cycle_PE_v6.1.xml $outgoing_runfolder_path/`;
  `touch $outgoing_runfolder_path/Data/C1-18-Firecrest1.3.4_09-07-2009_auto/Bustard1.3.4_09-07-2009_auto/GERALD_09-07-2009_auto/touch_file`;
  return 1;
}

sub set_rta_staging_analysis_area {
  my ($self, $indexed, $no_qseq_customs) = @_;
  $self->remove_staging();
  my $analysis_runfolder_path = $self->temp_directory() . $ANALYSIS_RUNFOLDER_PATH;
  my $bustard_path = qq{$analysis_runfolder_path/Data/Intensities/Bustard1.5.1_09-07-2009_RTA};
  my $recalibrated_path = qq{$bustard_path/GERALD_09-07-2009_RTA};
  `mkdir -p $recalibrated_path`;
  `mkdir $analysis_runfolder_path/Config`;
  `mkdir $analysis_runfolder_path/t`;
  `touch $bustard_path/s_1_1_001_qseq.txt`;
  `touch $bustard_path/s_1_2_001_qseq.txt`;
  `touch $bustard_path/s_2_1_001_qseq.txt`;
  `touch $bustard_path/s_2_2_001_qseq.txt`;
  `touch $bustard_path/s_1_1_001_qval.txt.gz`;
  `touch $bustard_path/s_1_2_001_qval.txt.gz`;
  `touch $bustard_path/s_2_1_001_qval.txt.gz`;
  `touch $bustard_path/s_2_2_001_qval.txt.gz`;
  `cp t/data/runfolder/Data/Intensities/Bustard_RTA/config.xml $bustard_path/`;
  `cp t/data/summary_files/after_v7_mp_hack_Summary.xml $recalibrated_path/Summary.xml`;
  `cp t/data/summary_files/after_v7_mp_hack_Summary.htm $recalibrated_path/Summary.htm`;
  `cp t/data/summary_files/Summary.xsl $recalibrated_path/Summary.xsl`;
  `cp t/data/runfolder/Data/Intensities/Bustard_RTA/GERALD/config.xml $recalibrated_path/`;
  if (!$no_qseq_customs) {
    foreach my $lane (1..8) {
      `touch $recalibrated_path/s_${lane}_1_001_qseq_custom.txt`;
      `touch $recalibrated_path/s_${lane}_2_001_qseq_custom.txt`;
    }
  }
  if ($indexed) {
    `cp t/data/Recipes/Recipe_GA2-PEM_MP_2x76Cycle+8_v7.7.xml $analysis_runfolder_path/`;
  } else {
    `cp t/data/Recipes/Recipe_GA2_37Cycle_PE_v6.1.xml $analysis_runfolder_path/`;
    `cp t/data/runfolder/Data/Intensities/single_end_3cycle_1tile_config.xml $analysis_runfolder_path/Data/Intensities/config.xml`;
    foreach my $lane ( 1..8 ) {
      foreach my $cycle ( 1..3 ) {
        qx{mkdir -p $analysis_runfolder_path/Data/Intensities/L00$lane/C$cycle.1};
        qx{cp t/data/runfolder/Data/Intensities/demo_cif.cif $analysis_runfolder_path/Data/Intensities/L00$lane/C$cycle.1/s_${lane}_1.cif};
      }
    }
  }
  `cp t/data/Recipes/TileLayout.xml $analysis_runfolder_path/Config/`;
  `touch $recalibrated_path/touch_file`;
  return {bustard_path => $bustard_path, recalibrated_path => $recalibrated_path, runfolder_path => $analysis_runfolder_path};
}

sub remove_staging {
  my ($self) = @_;
  my $staging = $self->temp_directory() . $NFS_STAGING_DISK;
  `rm -rf $staging`;
  return 1;
}

sub multiplexed_rta_pre_gerald {
  my ($self) = @_;
  my $paths = $self->set_rta_staging_analysis_area(1,1);
  my $bustard_path = $paths->{bustard_path};
  my $runfolder_path = $paths->{runfolder_path};
  `touch $bustard_path/s_1_3_001_qseq.txt`;
  `touch $bustard_path/s_2_3_001_qseq.txt`;
  `touch $bustard_path/s_1_3_001_qval.txt.gz`;
  `touch $bustard_path/s_2_3_001_qval.txt.gz`;
  foreach my $lane (3..7) {
    `touch $bustard_path/s_${lane}_1_0001_qseq.txt`;
    `touch $bustard_path/s_${lane}_2_0001_qseq.txt`;
    `touch $bustard_path/s_${lane}_3_0001_qseq.txt`;
    `touch $bustard_path/s_${lane}_1_0001_qval.txt.gz`;
    `touch $bustard_path/s_${lane}_2_0001_qval.txt.gz`;
    `touch $bustard_path/s_${lane}_3_0001_qval.txt.gz`;
  }
  `touch $bustard_path/s_8_1_0001_qval.txt.gz`;
  `touch $bustard_path/s_8_2_0001_qval.txt.gz`;
  `touch $bustard_path/s_8_3_0001_qval.txt.gz`;
  `cp t/data/mock_qseqs/1234/s_8_*qseq.txt $bustard_path/`;
  `cp t/barcodes/sanger.tags $runfolder_path/tag_set`;
  `cp t/data/runfolder/Data/Intensities/indexed_76cycle_config.xml $runfolder_path/Data/Intensities/config.xml`;
  return 1;
}

sub hiseq_runfolder_path {
  my ( $self, $mismatched_cycles ) = @_;

  $self->remove_staging();
  my $analysis_runfolder_path = $self->analysis_runfolder_path() . q{_A_80B5TABXX};
  my $intensity_path = $analysis_runfolder_path . q{/Data/Intensities};
  my $basecall_path = $intensity_path . q{/BaseCalls};

  qx{mkdir -p $basecall_path};

  qx{cp t/data/fix_configs/data_intensities_config.xml $intensity_path/config.xml};
  if ( $mismatched_cycles ) {
    qx{cp t/data/fix_configs/data_intensities_basecalls_diff_last_cycle_config.xml $basecall_path/config.xml};
  } else {
    qx{cp t/data/fix_configs/data_intensities_basecalls_config.xml $basecall_path/config.xml};
  }
  return $analysis_runfolder_path;
}

# for dropping the generated temporary part from paths
# and also anything which has the cwd in it, will be stripped out
# since this will not be stable between test runs
sub drop_temp_part_from_paths {
  my ( $self, $path ) = @_;
  my $temp_dir = $self->temp_directory();
  my $cwd = $self->cwd();
  $path =~ s{$temp_dir}{}gxms;
  $path =~ s{$cwd/}{}gxms;
  $path =~ s{$cwd}{}gxms;
  return $path;
}

### for ensuring that we pick up the correct conf_path with only one place to pick it up from
sub conf_path {
  my ( $self ) = @_;
  return $self->cwd() . q{/data/config_files};
}

# ensure that the environment variables do not get passed around and that extraneous files do not get left behind
sub DEMOLISH {
  my ( $self ) = @_;
  $ENV{ npg::api::request->cache_dir_var_name() } = q{};
  `rm -f *out; rm -f Latest_Summary`;
}

1;
